#include <linux/module.h>
#include <linux/kernel.h>
#include <rtl_fifo.h>
#include <rtl_time.h>
#include <rtl_sched.h>
#include <asm/io.h>
#include <linux/cons.h>
#include "common.h"

int number_of_tests=50;
int period=1000000;
int periodic_mode=0;
int fifo_size=4000;

MODULE_PARM(period,"i");
MODULE_PARM(number_of_tests,"i");
MODULE_PARM(periodic_mode,"i");

pthread_t thread;
timer_t timer;


void *thread_code(void *t) {

	struct timespec expected;
	struct timespec diff;
	struct timespec now;
	struct timespec min_diff;
	struct timespec max_diff;
	struct sample samp;
	struct itimerspec spec;
	int i;
	int cnt = 0;
	DECLARE_CPUID(cpu_id);

	rtl_printf ("Measurement task starts on CPU %d\n", cpu_id);
	rtl_gettime( &spec.it_value);	
	spec.it_value.tv_sec += 3;
	spec.it_interval = timespec_from_ns (period);
	rtl_printf ("Measurement task period %ld\n", spec.it_interval.tv_nsec);
	pthread_setperiod_np(thread,&spec);
	expected = spec.it_value;

	do {
		min_diff = timespec_from_ns(2000000000);
		max_diff = timespec_from_ns(-2000000000);

		for (i = 0; i < number_of_tests; i++) {
			++cnt;
			pthread_wait_np();

			rtl_gettime (&now);
			diff = now;
			timespec_sub (&diff, &expected);

			timespec_add (&expected, &spec.it_interval);

			if (timespec_lt (&diff, &min_diff)) {
				min_diff = diff;
			}
			if (timespec_gt (&diff, &max_diff)) {
				max_diff = diff;
			}

		}

		samp.min = min_diff;
		samp.max = max_diff;
		rtf_put(0, &samp, sizeof(samp));
/* 		rtl_printf("rmin: %11d, rmax: %11d\n", (int) (timespec_to_ns(&samp.min)), (int) (timespec_to_ns(&samp.max))); */
	} while (1);
	return 0;
}



int init_module(void)
{
	pthread_attr_t attr;
	struct rtl_sched_param sched_param;
	int thread_status;
        int fifo_status;

	rtf_destroy(0);
	fifo_status = rtf_create(0, fifo_size);
	if ((fifo_status != fifo_size)) {
		printk("RTL measurement test fail. fifo_status=%d\n",fifo_status);
		return -1;
	}


	printk("RTL measurement module on CPU %d\n",rtl_get_cpu_id());
	pthread_attr_init (&attr);
	pthread_attr_setcpu(&attr, 0);
	sched_param.sched_priority = 1;
	pthread_attr_setschedparam (&attr, &sched_param);
	printk("About to thread create\n");
	thread_status = pthread_create (&thread,  &attr, thread_code, (void *)1);
	if (thread_status < 0) {
		printk("failed to create RT-thread\n");
		return -1;
	} else {
		printk("created RT-thread\n");
	}

	return 0;
}


void cleanup_module(void)
{
        printk ("Removing module on CPU %d\n", rtl_get_cpu_id());
	pthread_delete_np (thread);
/* should be:	pthread_cancel(thread); pthread_join(thread); */
/* or:  	rtl_pthread_kill_other_threads(); */
	rtf_destroy(0);
}

