#include <linux/module.h>
#include <linux/kernel.h>
#include <rtl_fifo.h>
#include <rtl_time.h>
#include <rtl_sched.h>
#include <asm/io.h>
#include <linux/cons.h>
#include "common.h"

int number_of_tests=50;
int period=20000000;
int periodic_mode=0;
int fifo_size=4000;

pthread_t thread;

/* For testing purposes, if we are on a smp machine, try to 
  run the measurement module on the _other_cpu 
  on a uniprocessor, ignore this test.
*/
  
void *thread_code(void *t) {

	unsigned long last_time;
	unsigned long diff;
	unsigned long now;
	unsigned long min_diff;
	unsigned long max_diff;
	struct sample samp;
	struct itimerspec spec;
	int i;

	printk("Measurement task starts\n");
	rtl_gettime( &spec.it_value);	
	spec.it_interval = timespec_from_ns (period);
	printk("Measurement task period %d  %d \n",(int)spec.it_interval.tv_nsec, (int)spec.it_value.tv_nsec);
	pthread_setperiod_np(thread,&spec);
	do {
		min_diff = -1;
		max_diff = 0;
		rdtscl(last_time);
		for (i = 0; i < number_of_tests; i++) {
			pthread_wait_np();
			rdtscl(now);
			diff = (last_time< now? now- last_time : 0); //ignore overflows
			if(diff > max_diff)max_diff=diff;
			if(diff & (diff < min_diff)) min_diff = diff;
			last_time  = now;
		}

		samp.min.tv_sec = 0;
		samp.max.tv_sec = 0;
		samp.min.tv_nsec = min_diff;
		samp.max.tv_nsec = max_diff;
		rtf_put(0, &samp, sizeof(samp));
	} while (1);
	return 0;
}



int init_module(void)
{
	pthread_attr_t attr;
	struct rtl_sched_param sched_param;
	int thread_status;
        int fifo_status;

	rtf_destroy(0);
	fifo_status = rtf_create(0, fifo_size);
	if ((fifo_status != fifo_size)) {
		printk("RTL measurement test fail. fifo_status=%d\n",fifo_status);
		return -1;
	}


	printk("RTL measurement module on CPU %d\n",rtl_get_cpu_id());
	pthread_attr_init (&attr);
	sched_param.sched_priority = 1;
	pthread_attr_setschedparam (&attr, &sched_param);
	printk("About to thread create\n");
	thread_status = pthread_create (&thread,  &attr, thread_code, (void *)1);
	if (thread_status < 0) {
		printk("failed to create RT-thread\n");
		return -1;
	} else {
		printk("created RT-thread\n");
	}

	return 0;
}


void cleanup_module(void)
{
        printk ("Removing module on CPU %d\n", rtl_get_cpu_id());
	pthread_delete_np (thread);
/* should be:	pthread_cancel(thread); pthread_join(thread); */
/* or:  	rtl_pthread_kill_other_threads(); */
	rtf_destroy(0);
}

