#ifndef __RTL_COMPATIBILITY__
#define __RTL_COMPATIBILITY__

/* compatibility */

#include <linux/malloc.h>
#include <rtl_conf.h>
#include <asm/rt_time.h>

#define RT_TICKS_PER_SEC 1193180LL
#define RT_TIME_END 0x7fffFfffFfffFfffLL

extern long scaler_8254_to_ns;
extern long scaler_ns_to_8254;
extern inline struct timespec RTIME_to_timespec (RTIME t)
{
	long long ns;
	__asm__(
		"shld $10, %%eax, %%edx\n\t"
		"shl $10, %%eax\n\t"
		"mov %%edx, %%ecx\n\t"
		"mul %%ebx\n\t"  	/* multiply the low 32 bits of the counter by the scaler */
		"mov %%ecx, %%eax\n\t"
		"mov %%edx, %%ecx\n\t"	/* save the high 32 bits of the product */
		"mul %%ebx\n\t" 	/* now the high 32 bits */
		"add %%ecx, %%eax\n\t"
		"adc $0, %%edx\n\t"
		:"=A" (ns) : "b" (scaler_8254_to_ns), "A" (t): "cx");
	return timespec_from_ns(ns);
}

extern inline RTIME timespec_to_RTIME (struct timespec ts)
{
	long long t = timespec_to_ns (&ts);
	__asm__("mov %%edx, %%ecx\n\t"
		"mull %%ebx\n\t"
		"mov %%ecx, %%eax\n\t"
		"mov %%edx, %%ecx\n\t"	/* save the high 32 bits of the product */
		"mul %%ebx\n\t" 	/* now the high 32 bits */
		"add %%ecx, %%eax\n\t"
		"adc $0, %%edx\n\t"
		:"=A" (t) :"A" (t), "b" (scaler_ns_to_8254)
		:"cx");
	return t;
}


extern inline RTIME rt_get_time(void)
{
	struct timespec ts;
	rtl_gettime(&ts);
	return timespec_to_RTIME(ts);
}

extern inline RTIME rtl_set_periodic_mode (RTIME period)
{
	struct timespec ts;
	int mode_success;
	schedule_t *s = LOCAL_SCHED;
	ts = RTIME_to_timespec (period);
	mode_success = rtl_clock_set_periodic_mode (s->clock, &ts);
	if (mode_success == 0) {
		clock_gettime(s->clock, &ts); /* right now this is CLOCK_REALTIME regardless of s->clock */
		return timespec_to_ns(&ts);
	} else {
		return RT_TIME_END;
	}
}

extern inline int rtl_set_oneshot_mode (void)
{
	schedule_t *s = LOCAL_SCHED;
	return rtl_clock_set_oneshot_mode(s->clock);
}



#ifdef CONFIG_RTL_FP_SUPPORT
extern inline int rt_task_use_fp (pthread_t *task, int flag) {
	return pthread_use_fp_np (*task, flag);
}
#endif

/* cross-CPU-allowed functions */
extern int rt_task_suspend(pthread_t *task);
extern int rt_task_wakeup(pthread_t *thread);


extern inline int rt_task_delete(pthread_t *thread)
{
	int ret;
	ret = pthread_delete_np(*thread);
	return 0;
}

extern int rt_task_make_periodic (pthread_t *threadptr, RTIME start_time, RTIME period);

/* end of cross-CPU-allowed functions */


/* for periodic tasks -- wait until the next period */
extern inline int rt_task_wait (void)
{
	pthread_wait_np();
	return 0;
}


#define RT_TASK pthread_t


extern int rt_task_init (RT_TASK *t, void (*fn)(int data), int data, int stack_size, int priority);

#define rtl_init (a,b,c,d,e) { rt_task_init(a,b,c,d,e); }
#define rtl_fpinit (a,b,c,d,e) { rt_task_init(a,b,c,d,e); rtl_task_use_fp (a, 1); }

#endif
