/*
 * rtl_time.c
 *
 * architecture-dependent clock support
 *
 * Written by Michael Barabanov
 * Copyright VJY Associates 1998-1999
 * Released under the terms of the LGPL
 *
 */

#define RTL_TIME_DEBUG

#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/smp.h>
#include <asm/io.h>
#include <linux/errno.h>
#include <asm/system.h>
#include <linux/rtl.h>
#include <linux/sched.h>
#include <linux/timex.h>
#include <linux/cons.h>
#include <linux/mc146818rtc.h>

#define RTL_DEBUG
#include <rtl_debug.h>

#include <rtl_sync.h>
#include <asm/rtl_time.h>

struct hw_clock i8254 /* __attribute__ ((aligned (64))) */;
#ifdef __SMP__
struct hw_clock apic_clock[NR_CPUS] /*__attribute__ ((aligned (64)))*/;
#endif

static void default_handler( struct pt_regs *regs)
{
}

int rtl_clock_sethandler (struct hw_clock *h, RTL_CLOCK_HANDLER fn)
{
	if (h -> handler != &default_handler) {
		return -EBUSY;
	}
	h->handler = fn;
	return 0;
}

int rtl_clock_unsethandler (struct hw_clock *h)
{
	if (h->handler == &default_handler) {
		return -EINVAL;
	}
	h-> handler = &default_handler;
	return 0;
}

int rtl_clock_ishandlerset (struct hw_clock *h)
{
	return h-> handler != &default_handler;
}


/* 8254  */
static int _8254_init(struct hw_clock *);
static void _8254_uninit(struct hw_clock *);
static int _8254_settimer(int , struct timespec *);
static void _8254_checklinuxirq(void);
int rtl_gettime (struct timespec *);
static void grab_8254(void);

struct hw_clock i8254_data =  {
	_8254_init,
	_8254_uninit,
	rtl_gettime,
	_8254_settimer,
	_8254_checklinuxirq,&default_handler,0,CLOCK_MODE_UNINITIALIZED
};

extern void (*kd_mksound)(unsigned int hz, unsigned int ticks);

static long long (*rtl_do_get_time)(void);

static long LATCH_NS;
static struct timespec LATCH_TIMESPEC;
static struct timespec MAX_LATCH_ONESHOT;

/* getting global time from 8254 */

#define READ_CNT0(var) \
{ var = inb(0x40); var |= (inb(0x40) << 8); }

#define READ_CNT2(var) \
{ var = inb(0x42); var |= (inb(0x42) << 8); }


#define LATCH_CNT0() \
outb(0xd2, 0x43);

#define LATCH_CNT0_AND_2() \
outb(0xda, 0x43);

#define LATCH_CNT2() \
outb(0xd8, 0x43);

#define WRITE_COUNTER_ZERO16(x)  { \
	outb_p(x&0xff,0x40); outb_p((x>>8)&0xff,0x40);\
      	clock_counter =x; \
} 

#define WRITE_COUNTER_ZERO8(x)  { \
	outb_p(x&0xff,0x40); \
	clock_counter =x; \
}

#define WRITE_COUNTER_ZERO_ONESHOT(x) { \
	if (x <= 0xff) { \
		if (max_counter0 != 0xff) { \
			outb_p(0x10, 0x43);       /* 8254, channel 0, mode 0, lsb */ \
			max_counter0 = 0xff; \
		} \
		WRITE_COUNTER_ZERO8(x); \
	} else { \
		if (max_counter0 != 0xffff) { \
			outb_p(0x30, 0x43);        /* 8254, channel 0, mode 0, lsb+msb */ \
			max_counter0 = 0xffff; \
		} \
		WRITE_COUNTER_ZERO16(x); \
	} \
}

static volatile long long rt_time; /* last time (in 8254 ticks) when counter 2 was equal to last_c2 */
static volatile int last_c2;
unsigned long scaler_8254_to_ns;  /* =8380965  ns*100 */
unsigned long scaler_ns_to_8254;
#define LATCH2 0x8000

#ifdef __SMP__
static spinlock_t lock8254;
/* static spinlock_t rtl_time_lock; */
#define RTL_SPIN_LOCK rtl_time_lock
#endif


long long global_8254_gettime (void)
{
	register int c2;
	int flags;
	long long t;

	rtl_spin_lock_irqsave (&lock8254, flags);
	LATCH_CNT2();
	READ_CNT2(c2);
	rt_time += ((c2 < last_c2) ? (last_c2 - c2) / 2 : (last_c2 - c2 + LATCH2) / 2);
	last_c2 = c2;
	__asm__(
		"shld $10, %%eax, %%edx\n\t"
		"shl $10, %%eax\n\t"
		"mov %%edx, %%ecx\n\t"
		"mul %%ebx\n\t"  	/* multiply the low 32 bits of the counter by the scaler */
		"mov %%ecx, %%eax\n\t"
		"mov %%edx, %%ecx\n\t"	/* save the high 32 bits of the product */
		"mul %%ebx\n\t" 	/* now the high 32 bits */
		"add %%ecx, %%eax\n\t"
		"adc $0, %%edx\n\t"
		:"=A" (t) : "b" (scaler_8254_to_ns), "A" (rt_time): "cx");
	rtl_spin_unlock_irqrestore (&lock8254, flags);
	return t;
}


/* getting global time from Pentium TSC */
static unsigned long scaler_pentium_to_ns;
extern unsigned long fast_gettimeoffset_quotient;
long rtl_delay_mult;
int can_change_latch2;

int I8253_channel2_free(void)
{
	return can_change_latch2;
}


long long pent_gettime(void)
{

	long long t;

	/* time = counter * scaler_pentium_to_ns / 2^32 * 2^5; */
	/* Why 2^5? Because the slowest Pentiums run at 60 MHz */

	__asm__("rdtsc\n\t"
		"mov %%edx, %%ecx\n\t"
		"mul %%ebx\n\t"  	/* multiply the low 32 bits of the counter by the scaler_pentium */
		"mov %%ecx, %%eax\n\t"
		"mov %%edx, %%ecx\n\t"	/* save the high 32 bits of the product */
		"mul %%ebx\n\t" 	/* now the high 32 bits of the counter */
		"add %%ecx, %%eax\n\t"
		"adc $0, %%edx\n\t"
		"shld $5, %%eax, %%edx\n\t"
		"shl $5, %%eax\n\t"
		:"=A" (t) : "b" (scaler_pentium_to_ns) : "cx");
	return t;
}

/* TODO  I commented these out here and where used, Don't think we need it
   except on 486 ! Didn't want to debug */
//static unsigned long (*save_do_gettimeoffset)(void);
/* we can't allow Linux to read the clocks; it must use this function */
//static unsigned long do_rt_gettimeoffset(void)
//{
//	return 0; /* TODO: take code from 2.0 version */
//}


static void rtl_kd_nosound(unsigned long ignored)
{
	int flags;
	rtl_no_interrupts(flags);
	outb(inb_p(0x61) & 0xfd, 0x61);
	rtl_restore_interrupts(flags);
}

static void rtl_kd_mksound(unsigned int hz, unsigned int ticks)
{
	static struct timer_list sound_timer = { NULL, NULL, 0, 0,
						 rtl_kd_nosound };

	unsigned int count = 0;

	if (hz > 20 && hz < 32767)
		count = 1193180 / hz;
	
	cli();
	del_timer(&sound_timer);
	if (count) {
		int flags;
		rtl_spin_lock_irqsave (&lock8254, flags);
		outb_p(inb_p(0x61)|3, 0x61);
		if (can_change_latch2) {
			outb_p(0xB6, 0x43);
			outb_p(count & 0xff, 0x42);
			outb((count >> 8) & 0xff, 0x42);
		}
		rtl_spin_unlock_irqrestore (&lock8254, flags);

		if (ticks) {
			sound_timer.expires = jiffies+ticks;
			add_timer(&sound_timer);
		}
	} else
		rtl_kd_nosound(0);
	sti();
	return;
}


static void (*save_kd_mksound)(unsigned int hz, unsigned int ticks);

static void ungrab_8254(void){
//	do_gettimeoffset = save_do_gettimeoffset;
	rtl_kd_mksound(0, 0);
	kd_mksound = save_kd_mksound;
}

void grab_8254(void){
	int flags;

	kd_mksound(0, 0); /* clear the possibly pending sound timer */
	save_kd_mksound = kd_mksound;
	kd_mksound = rtl_kd_mksound;

	scaler_8254_to_ns = muldiv (NSECS_PER_SEC, 1 << 22, CLOCK_TICK_RATE);
	scaler_ns_to_8254 = muldiv (CLOCK_TICK_RATE, 1 << 31, NSECS_PER_SEC / 2);
	LATCH_NS = muldiv (LATCH, NSECS_PER_SEC, CLOCK_TICK_RATE);
	LATCH_TIMESPEC.tv_sec = 0;
	LATCH_TIMESPEC.tv_nsec = LATCH_NS;

	MAX_LATCH_ONESHOT.tv_sec = 0;
	MAX_LATCH_ONESHOT.tv_nsec = LATCH_NS * 3 / 4;

	rtl_no_interrupts(flags);
	if (boot_cpu_data.x86_capability & X86_FEATURE_TSC) {
		can_change_latch2 = 1;
		/* scaler_pentium ==  2^32 / (2^5 * (cpu clocks per ns)) */

		scaler_pentium_to_ns = muldiv (get_tsc_quotient(), 1000, 1 << 5);
		rtl_do_get_time = pent_gettime;
		
/* printk("scaler_pentium_to_ns = %d\n", (int) scaler_pentium_to_ns);  */
	} else {
		can_change_latch2 = 0;
		/* program channel 2 of the 8254 chip for periodic counting */
		outb_p(0xb6, 0x43);     /* binary, mode 3, LSB/MSB, ch 2 */
		outb_p(LATCH2 & 0xff, 0x42);
		outb_p((LATCH2 >> 8) & 0xff, 0x42);
		outb_p((inb_p(0x61) & 0xfd) | 1, 0x61); /* shut up the speaker and enable counting */

		LATCH_CNT2();
		READ_CNT2(last_c2);
		rt_time = 0;
		rtl_do_get_time = global_8254_gettime;
	}


	/* TODO What's going on here ?  I don't think we need this VY*/
//	save_do_gettimeoffset = do_gettimeoffset;
//	do_gettimeoffset = do_rt_gettimeoffset;

	/*
	current_time.tv_sec = xtime.tv_sec;
	current_time.tv_nsec = 0;

	rtl_time_offset = 0;
	rtl_time_offset = rtl_timespec_to_rtime(&current_time) - rtl_get_time();

	*/
	rtl_restore_interrupts(flags);
}

void rtl_clock_clear(struct hw_clock *h)
{
	h->uninit(h);
}


int rtl_gettime (struct timespec *ts)
{
	long long t;
       if(rtl_do_get_time) t=  rtl_do_get_time();
       else {
	       printk("can't call rtl_gettime without initializing a clock\n");
	       return 0;
       }
	__asm__("divl %%ecx\n\t"
		:"=a" (ts->tv_sec), "=d" (ts->tv_nsec)
		: "A" (t), "c" (NSECS_PER_SEC)
		:"cx");
	return 0;
}




/* the 8254 clock */

static unsigned int clock_counter; /* current latch value */
static int max_counter0;    /* 0xff or 0xffff - one or two byte counts */


static inline long RTIME_to_8254_ticks(long t)
{
	__asm__("mull %0"
		:"=d" (t)
		:"a" (t), "0" (scaler_ns_to_8254)
		:"ax");
		
	return (t);
}


extern void rtl_hard_enable_irq(int);
static unsigned int _8254_irq(unsigned int irq, struct pt_regs *regs)
{
	int flags;
	rtl_spin_lock_irqsave (&lock8254, flags);
	if (i8254.mode == CLOCK_MODE_PERIODIC) {
		timespec_add (&i8254.time, &i8254.resolution);
	} else {
		i8254.istimerset = 0;
	}
	rtl_hard_enable_irq(0);
	rtl_spin_unlock_irqrestore (&lock8254, flags);
	i8254.handler(regs);
	rtl_spin_lock_irqsave (&lock8254, flags);
	_8254_checklinuxirq();
	rtl_spin_unlock_irqrestore (&lock8254, flags);
	return 0;
}


static int _8254_settimer (int mode, struct timespec *interval)
{
	int flags;
	long t;
	struct timespec ts;

	rtl_spin_lock_irqsave (&lock8254, flags);
	if (mode == CLOCK_MODE_PERIODIC && interval) {
		if (i8254.mode != mode) {
			outb_p(0x34,0x43); /* binary, mode 2, LSB/MSB, ch 0 */
			i8254.mode = mode;
		}
		t = RTIME_to_8254_ticks (interval -> tv_nsec);
		if(t < 10){ t = LATCH; printk("RTL settimer set too low, ignoring\n");}
		WRITE_COUNTER_ZERO16 (t);

		rtl_gettime(&i8254.time);
		i8254.resolution.tv_sec = 0;
		i8254.resolution.tv_nsec = muldiv (t, NSECS_PER_SEC, CLOCK_TICK_RATE);
		i8254.istimerset = 1;
	} else if (mode == CLOCK_MODE_ONESHOT) {
		if (i8254.mode != mode) {
			outb_p(0x30, 0x43);       /* 8254, channel 0, mode 0, lsb+msb */
			max_counter0 = 0xffff;
			i8254.mode = mode;
			i8254.resolution.tv_sec = 0;
			i8254.resolution.tv_nsec = 1;
		}

		if (interval) {
			ts = *interval;
			if (timespec_gt(&ts, &MAX_LATCH_ONESHOT)) {
				ts = MAX_LATCH_ONESHOT;
			} else if (timespec_lt(&ts, &i8254.resolution)) {
				ts = i8254.resolution;
			}
		} else {
			ts = MAX_LATCH_ONESHOT;
		}

		t = RTIME_to_8254_ticks(ts.tv_nsec) + 1; /* FIXME (+1) */
		WRITE_COUNTER_ZERO_ONESHOT(t);
		i8254.istimerset = 1;
	}
	rtl_spin_unlock_irqrestore(&lock8254, flags);
	return 0;
}


/* #define __RTL_DEBUG_TIME__ */

void _8254_checklinuxirq (void)
{
/*	if (test_and_set_bit(0, &checklinuxirq[clock])) {
		return;
	}
	*/
	struct timespec ts;

	rtl_gettime(&ts);
	if  (timespec_ge (&ts, &i8254.linux_time)) {
		timespec_add (&i8254.linux_time, &LATCH_TIMESPEC);
		rtl_global_pend_irq (0);
		if  (timespec_ge (&ts, &i8254.linux_time)) {
#ifdef __RTL_DEBUG_TIME__
			conpr("RTL: lost jiffies\n");
#endif
			i8254.linux_time = ts;
			timespec_add (&i8254.linux_time, &MAX_LATCH_ONESHOT);
		}
	}

	if (!i8254.istimerset) {
		_8254_settimer(CLOCK_MODE_ONESHOT, &MAX_LATCH_ONESHOT);
	}
/* 	clear_bit (0, &checklinuxirq [clock]); */
}


static int _8254_init (struct hw_clock *clock)
{
	int flags;
	rtl_no_interrupts (flags);
	rtl_gettime(&i8254.linux_time);
	timespec_add (&i8254.linux_time, &LATCH_TIMESPEC);
	rtl_request_global_irq(0, _8254_irq);
	_8254_settimer (CLOCK_MODE_ONESHOT, &LATCH_TIMESPEC);
	rtl_restore_interrupts (flags);
	return 0;
}

static void _8254_uninit (struct hw_clock *clock)
{
	int flags;
	if (clock -> mode == CLOCK_MODE_UNINITIALIZED) {
		return;
	}
	clock->handler = &default_handler;
	rtl_spin_lock_irqsave (&lock8254, flags);
	outb_p(0x34,0x43);		/* binary, mode 2, LSB/MSB, ch 0 */
	WRITE_COUNTER_ZERO16(LATCH);
	rtl_free_global_irq(0);
	clock -> mode = CLOCK_MODE_UNINITIALIZED;
	rtl_spin_unlock_irqrestore (&lock8254, flags);
}



#ifdef __SMP__
/* APIC clocks */

static int apic_clock_init(struct hw_clock *clock);
static void apic_clock_uninit(struct hw_clock *clock);
static void apic_checklinuxirq(void);
static int apic_settimer(int , struct timespec *);

struct hw_clock apic_clock_structure = 
{apic_clock_init, apic_clock_uninit, rtl_gettime, apic_settimer,
	       apic_checklinuxirq, &default_handler,0,CLOCK_MODE_UNINITIALIZED};

/* struct hw_clock apic_clock[NR_CPUS]; */

#define LOCAL_TIMER_VECTOR	0x41

static unsigned long scaler_ns_to_apic;

static inline long RTIME_to_apic_ticks(long t)
{
	__asm__("mull %0"
		:"=d" (t)
		:"a" (t), "0" (scaler_ns_to_apic)
		:"ax");
		
	return (t);
}


static inline int rtl_apic_write_initial_count (long count)
{
	unsigned int tmp_value;
	tmp_value = apic_read(APIC_TMICT);
	apic_write (APIC_TMICT, count);
	return 0;
}

/* this is defined in arch/i386/kernel/smp.c */
#define APIC_DIVISOR 16

static int apic_settimer (int mode, struct timespec *interval)
{
	int flags;
	long t;
	struct timespec ts;
	unsigned long lvtt1_value;
	unsigned int tmp_value;
	DECLARE_CPUID(cpu_id);
	struct hw_clock *apic = &apic_clock [cpu_id];

	rtl_no_interrupts (flags);
	if (mode == CLOCK_MODE_PERIODIC && interval) {
		if (apic->mode != mode) {
			apic -> mode = mode;
			tmp_value = apic_read(APIC_LVTT);
			lvtt1_value = APIC_LVT_TIMER_PERIODIC | LOCAL_TIMER_VECTOR;
			apic_write(APIC_LVTT , lvtt1_value);
		}
		t = RTIME_to_apic_ticks(interval->tv_nsec);
		rtl_apic_write_initial_count (t);
		rtl_gettime (&apic->time);
		apic -> resolution = *interval;
		apic->istimerset = 1;
	} else if (mode == CLOCK_MODE_ONESHOT) {
		if (apic->mode != mode) {
			apic -> mode = mode;
			apic -> resolution.tv_sec = 0;
			apic -> resolution.tv_nsec = 1;

			tmp_value = apic_read(APIC_LVTT);
			lvtt1_value = LOCAL_TIMER_VECTOR;
			apic_write(APIC_LVTT , lvtt1_value);

		}

		if (interval) {
			ts = *interval;
			if (timespec_gt(&ts, &MAX_LATCH_ONESHOT)) {
				ts = MAX_LATCH_ONESHOT;
			} else if (timespec_lt(&ts, &apic->resolution)) {
				ts = apic->resolution;
			}
		} else {
			ts = MAX_LATCH_ONESHOT;
		}


		t = RTIME_to_apic_ticks(ts.tv_nsec) + 1; /*FIXME (+1)*/
		rtl_apic_write_initial_count (t);
		apic->istimerset = 1;
	}

	rtl_restore_interrupts(flags);
	return 0;
}

static void apic_checklinuxirq (void)
{
	DECLARE_CPUID(cpu_id);
	struct hw_clock *apic = &apic_clock[cpu_id];
	struct timespec ts;
	rtl_gettime (&ts);
	if  (timespec_ge (&ts, &apic->linux_time)) {
		rtl_local_pend_vec (LOCAL_TIMER_VECTOR,cpu_id);
		timespec_add (&apic->linux_time, &LATCH_TIMESPEC);
		if  (timespec_ge (&ts, &apic->linux_time)) {
#ifdef __RTL_DEBUG_TIME__
			conpr("RTL: lost linux APIC timer interrupts\n");
#endif
			apic->linux_time = ts;
			timespec_add (&apic->linux_time, &MAX_LATCH_ONESHOT);
		}
	}
	if (!apic -> istimerset) {
		apic_settimer (CLOCK_MODE_ONESHOT, &MAX_LATCH_ONESHOT);
	}
}

static unsigned int apic_timer_irq(struct pt_regs *r)
{
	DECLARE_CPUID(cpu_id);
	struct hw_clock *apic = &apic_clock[cpu_id];
	if (apic->mode == CLOCK_MODE_PERIODIC) {
		timespec_add (&apic->time, &apic->resolution);
	} else {
		apic->istimerset = 0;
	}
	apic_checklinuxirq();
	apic -> handler (r);
	return 0;
}


int volatile apic_init_flag;

static unsigned int apic_init_irq(struct pt_regs *r)
{
	DECLARE_CPUID(cpu_id);
	apic_settimer (CLOCK_MODE_ONESHOT, NULL);
	rtl_free_local_irq (LOCAL_TIMER_VECTOR, cpu_id);
	clear_bit (0, &apic_init_flag);
	return 0;
}

/* apics are initialized to the oneshot mode by default */
static int apic_clock_init (struct hw_clock *clk)
{
	int flags;
	DECLARE_CPUID(cpu_id);

	scaler_ns_to_apic = muldiv (muldiv (get_apic_calibration_result(), 1 << 31, NSECS_PER_SEC / 2), CLOCK_TICK_RATE, LATCH * APIC_DIVISOR);

	rtl_no_interrupts (flags);
	rtl_gettime(&clk->linux_time);
	timespec_add (&clk->linux_time, &LATCH_TIMESPEC);

	if (clk->apic_cpu == cpu_id) {
		apic_settimer (CLOCK_MODE_ONESHOT, NULL);
	} else {
		set_bit (0, &apic_init_flag);
		rtl_request_local_irq (LOCAL_TIMER_VECTOR, apic_init_irq, clk->apic_cpu);
		while (test_bit(0, &apic_init_flag));
	}
	rtl_request_local_irq (LOCAL_TIMER_VECTOR, apic_timer_irq, clk->apic_cpu);
	rtl_restore_interrupts (flags);
	return 0;
}


static unsigned int apic_uninit_irq(struct pt_regs *r)
{
	DECLARE_CPUID(cpu_id);
	apic_settimer (CLOCK_MODE_PERIODIC, &LATCH_TIMESPEC);
/* 	rtl_free_local_irq (LOCAL_TIMER_VECTOR, cpu_id); */
	free_ipi (cpu_id);
	clear_bit (0, &apic_init_flag);
	return 0;
}

static void apic_clock_uninit (struct hw_clock *clock)
{
	int flags;
	DECLARE_CPUID(cpu_id);

	if (clock -> mode == CLOCK_MODE_UNINITIALIZED) {
		return;
	}

	if (clock->apic_cpu == cpu_id) {
		rtl_no_interrupts (flags);
		clock->handler = &default_handler;
		apic_settimer (CLOCK_MODE_PERIODIC, &LATCH_TIMESPEC);
		clock -> mode = CLOCK_MODE_UNINITIALIZED;

		rtl_free_local_irq (LOCAL_TIMER_VECTOR, cpu_id);
		rtl_restore_interrupts (flags);
	} else {
#if 1
		rtl_no_interrupts (flags);
		clock->handler = &default_handler;

		set_bit (0, &apic_init_flag);
		rtl_free_local_irq (LOCAL_TIMER_VECTOR, clock->apic_cpu);
		request_ipi (apic_uninit_irq, clock->apic_cpu);
		rtl_reschedule(clock->apic_cpu);
/* 		rtl_request_local_irq (LOCAL_TIMER_VECTOR, apic_uninit_irq, clock - CLOCK_APIC); */
		rtl_restore_interrupts (flags);
		while (test_bit(0, &apic_init_flag)); /* wait for apic to uninit */
		clock -> mode = CLOCK_MODE_UNINITIALIZED;
#endif
	}

}


#endif




/* returns a pointer to the clock structure of the controlling hw clock */
struct hw_clock * rtl_best_clock(unsigned int cpu_id,RTL_CLOCK_HANDLER f)
{
	struct hw_clock *theclock;
#ifdef __SMP__
	if (smp_found_config) {
		theclock = &apic_clock[cpu_id];
	} else {
		theclock = &i8254;
	}
#else
	theclock = &i8254;
#endif
	if (rtl_clock_sethandler (theclock, f) == 0) {
		return theclock;
	} else {
		return 0;
	}
}


spinlock_t rtl_tqueue_lock;
int init_module (void)
{
	spin_lock_init (&lock8254);
	spin_lock_init (&rtl_tqueue_lock);
	rtl_debug = 0;

	rtl_delay_mult = muldiv (loops_per_sec, 1 << 31, NSECS_PER_SEC / 2);

	grab_8254();
	i8254 = i8254_data;
#ifdef __SMP__
	{
	int i;	
		for (i = 0; i < rtl_num_cpus(); i++) {
			int cpu = cpu_logical_map (i);
			apic_clock[cpu] = apic_clock_structure;
			apic_clock[cpu].apic_cpu = cpu;
		}
	}
#endif
	return 0;
}

void cleanup_module(void)
{
#ifdef __SMP__
	if (smp_found_config) {
		int i;
		for (i = 0; i < rtl_num_cpus(); i++) {
			int cpu = cpu_logical_map(i);
			apic_clock_uninit (&apic_clock[cpu]);
		}
	}
#endif
	_8254_uninit(&i8254);
	ungrab_8254();
}

/* the following should really belong to rtl_core */

/* compatibility irq handler table */
#include <asm/rt_irq.h>
RTL_V1_HANDLER rtl_v1_irq[NR_IRQS];

extern unsigned int rtl_compat_irq_handler(unsigned int irq, struct pt_regs *regs)
{
	rtl_v1_irq[irq]();
	rtl_hard_enable_irq(irq);
	return 0;
}


/* debugging */
unsigned int rtl_debug;

void rtl_setdebug (unsigned int state)
{
	rtl_debug = state;
}

unsigned int rtl_getdebug (void)
{
	return rtl_debug;
}

/* SMP rescheduling */
#ifdef __SMP__
#define RTL_RESCHEDULE_VECTOR 0x51
int request_ipi (unsigned int (*f)(struct pt_regs *r), int cpu)
{
	return rtl_request_local_irq (RTL_RESCHEDULE_VECTOR, f, cpu);
}

int free_ipi (int cpu)
{
	return rtl_free_local_irq (RTL_RESCHEDULE_VECTOR, cpu);
}
#endif
